;
; chrin.asm - CHRIN, GETIN kernal extension handler
;
; Written by
;  Kajtar Zsolt <soci@c64.rulez.org>
;
;    This file is part of IDEDOS the IDE64 disk operating system
;    See README for copyright notice.
;
;    This program is free software; you can redistribute it and/or modify
;    it under the terms of the GNU General Public License as published by
;    the Free Software Foundation; either version 2 of the License, or
;    (at your option) any later version.
;
;    This program is distributed in the hope that it will be useful,
;    but WITHOUT ANY WARRANTY; without even the implied warranty of
;    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;    GNU General Public License for more details.
;
;    You should have received a copy of the GNU General Public License
;    along with this program; if not, write to the Free Software
;    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
;
;**************************************************
; Chrin & getin
;
; Used registers: A (X Y preserved)
; Status: C (N Z for Accu)
; C=1:
; A=6 (not input file)
; Status response:
; ST=$00 if no error
; ST=$40 if last byte of file, or end of line on channel #15, or end of record.
; ST=$42 if no more bytes (open error, output file, device...)
; ST=$52 if no more bytes, because of readerror (open error, output file, device...)
; ST=$80 if device not present
;
;Note:
;Directory summ at end will be incorrect if multiple read at same time...
;And directory is half done anyway.
;**************************************************
;Buffer:
;
;current file position: (buffer_address_low_exp + buffer_address_hi_exp << 8 + 1) & 0x7ff | buffer_filepos2_exp <<8 ...
;if buffer_address_end_hi_exp == 0 then
;       buffer contains no data, have to read first
;       buffer_address_low/hi_exp must point before the buffer
;	buffer_address_end_low_exp must be 0
;
;if buffer_address_end_hi_exp is after the end of buffer then
;	the whole buffer is needed and there are more sectors to come
;
;if buffer_address_hi_exp & 0x80 then
;       file ends in this memory page
;       buffer_address_end_low_exp selects last byte of buffer
;
;buffer_address_low/hi_exp is only before buffer if it contains no data! so buffer_address_end_low/hi_exp=0
;buffer_address_low/hi_exp points to the previous byte read from buffer if any
;
; file position pointer does not advance if eof...
;--------------------------------------
; GETIN
;--------------------------------------
		.if (CHRIN_CHROUT_RAM=1) & (SUPERCPU=0)
fake		= newchrin_ram
fake2		= newgetin
		.fi

newgetin	.proc
		.if SUPERCPU=1
		phx
		.else
		stx help_x_exp
		.fi
		#ldx_l_a buffer_input_exp
		lda buffer_info_exp,x
		bpl newchrin.getin
back		cmp #INFO_KERNAL
		bne newchrin.chrin_not_data_sector
                .if SUPERCPU=1
                pla
                .fi
		#psr $f140
		.if SUPERCPU=1
                pha
                .fi
		cpx #STDIN
		bne nonstandard
		lda buffer_output_exp
		eor #STDOUT
		bne nonstandard
		.if SUPERCPU=1
                plx
                .fi
		#psr switch_norm
		#psr gotobank3		
                lda #$00
bbb             .if SUPERCPU=0
		ldx help_x_exp
                .fi
		ora #$00                ;Now we switch to bank3 to prevent
                jmp exit2basic          ;raster bugs when calling $ffe4
                                        ;instantly to scan keyboard...
nonstandard	lda buffer_drive_exp,x	
		.if SUPERCPU=1
                plx
                .fi
		cmp #4
                bcc bbb
                .if SUPERCPU=0
		ldx help_x_exp
                .fi
		pla
                pla
	        gcs newchrin.getinen
;--------------------------------------
; CHRIN
;--------------------------------------
newchrin	.proc
		.if SUPERCPU=1
		phx
		.else
		stx help_x_exp
		.fi
		#ldx_l_a buffer_input_exp
		lda buffer_info_exp,x
		bmi chrin_not_data_sector

;BLK_RW_DIRTY, BLK_READ, BLK_RW

getin		.if SUPERCPU=1
		lda $a6
		pha
		lda #1
		sta $a6			;SCPU Page 1
		clc
hard_chrin	lda buffer_address_low_exp,x
		inc a
		sta buffer_address_low_exp,x
		.else
hard_chrin	asr #$00
		isb buffer_address_low_exp,x
		eor #$ff
		.fi
		beq chrin_next_bank
		sta $a4
		lda buffer_address_hi_exp,x
		bmi chrin_data_sector_last
chrin_bank_nen2	sta $a5
chrin_bank_nend

		.if SUPERCPU=1		;Get byte
chrin_lastb
		lda [$a4]
		plx			;Restore $a6
		stx $a6
chrin_v		plx
		.else
chrin_lastb	ldx #0

		lda ($a4,x)
chrin_v		ldx help_x_exp
		.fi
		ora #$00		;Set NZ flags
		jmp exit2basic
;---------------------------------
chrin_next_bank	sta $a4
		.if SUPERCPU=1
		lda buffer_address_hi_exp,x
		inc a
		sta buffer_address_hi_exp,x
		.else
		isb buffer_address_hi_exp,x
		eor #$ff
		.fi
		cmp buffer_address_end_hi_exp,x
		bcc chrin_bank_nen2
		ora #$80
		sta buffer_address_hi_exp,x;Mark next bank

chrin_data_sector_last
		and #$7f
		sta $a5
		lda $a4
		cmp buffer_address_end_low_exp,x
		bcc chrin_bank_nend
		gcs chrin_next
;---------------------------------------

;BLK_WRITE_DIRTY, BLK_RW_EOF_DIRTY, BLK_R_EOF, BLK_RW_EOF, BLK_RW_EOF_FULL, BLK_WRITE_FULL, BLK_WRITE, STATUS, KERNAL, IGNORE

chrin_not_data_sector
		cmp #INFO_KERNAL
		bne nchrin_stand

		lda buffer_drive_exp,x
		cmp #4
		.if SUPERCPU=1
                plx
		.else
		ldx help_x_exp
                .fi
		bcc vew			;serial device?

getinen		lda Status
		beq sok
		lda #$0d
		clc
		gcc serr

sok		
		.if JIFFYDOS_ACCEL_CHRINOUT=1
		jsr OP_cfg16up
		jsr iecin
		jsr OP_cfg16down
                lda SerialInBuf
                jmp exit2basic
                .else

                .if ((JIFFYDOS_ACCEL=1) & (SUPERCPU=0)) | (DOLPHINDOS_ACCEL=1)
                lda dos_accel_internal_exp
                beq nx
                jsr OP_cfg16up
		jsr iecin
		jsr OP_cfg16down
                lda SerialInBuf
                jmp exit2basic
                .fi

nx              #psr $ee13
                jmp exit2basic
                .fi		;JIFFYDOS ACCEL CHRINOUT

vew		sta ActualInput
		ora #0
		bne bvb
                lda #CMOS_FUNCTION_KEYS
                and setup_atribut2_exp
                bne functionkeys

bvb		#psr $f157
serr		jmp exit2basic
;---------------------------------------

;BLK_WRITE_DIRTY, BLK_RW_EOF_DIRTY, BLK_R_EOF, BLK_RW_EOF, BLK_RW_EOF_FULL, BLK_WRITE_FULL, BLK_WRITE, STATUS, IGNORE

nchrin_stand	cmp #INFO_STATUS
		bne nchrin_status
                
		lda buffer_type_exp,x
                cmp #BUFFERMODE_PCLINK
                bne norm
                
                lda status_pclink_pos_exp
                bpl fe
                #phy_a
                jsr OP_save_zeropage_scpu01up
		jsr pclink_status
		#sty_l_a status_pclink_size_exp
                jsr OP_restore_zeropage_scpu01down
                #ply_a
                #lda0

fe              tax
		clc
                adc #1
		cmp status_pclink_size_exp
                bcc readb
                lda #$40
                sta Status
                asl
                
readb           sta status_pclink_pos_exp
		lda pclink_status_exp,x
                gcc chrin_v

norm		lda buffer_drive_exp,x
		sta current_drive_exp
                tax
		lda status_pos_exp,x
		bpl ok_st		;No pending request

		.if SUPERCPU=1
		phy
                tay
		.else
		tya
		pha
		ldy status_pos_exp,x
		.fi
		jsr OP_save_zeropage_scpu01up
		tya
		jsr LWOP_disk_status
		jsr OP_restore_zeropage_scpu01down
		#ldx_l_a current_drive_exp
		#ply_a

ok_st		lda status_pos_exp,x
		cmp status_size_exp,x
                bcs gon
		txa
		lsr
		ror			;*64
		ror
		ora status_pos_exp,x
		.if SUPERCPU=1
		pha
		and #$3f
		inc a
		sta status_pos_exp,x
		plx
		.else
		inc status_pos_exp,x
		tax
		.fi
		lda drive_status_exp,x
		gcc chrin_v		;end of message? (c=0)

gon     	
		#OP_scpu01up
		jsr LWOP_reporterror_ok
		lsr			;lda #$40
		sta Status
		#OP_scpu01down
		lda #$0d
					;clc
		gcc chrin_v
;---------------------------------------

;BLK_WRITE_DIRTY, BLK_RW_EOF_DIRTY, BLK_R_EOF, BLK_RW_EOF, BLK_RW_EOF_FULL, BLK_WRITE_FULL, BLK_WRITE, IGNORE

nchrin_status	lda #$42		;All the rest...
		sta Status
nchrin_err_end	clc
		.if SUPERCPU=1
		plx
		.else
		ldx help_x_exp
		.fi
		#lda0			;$c7
		jmp exit2basic
;---------------------------------------
chrin_next	#stx_l_a buffer_currentp_exp
		#phy_a
		jsr OP_save_zeropage_scpu01up

		jsr XIT_do_next
		bcc *+6
		lda #$52
		sta Status

		jsr OP_restore_zeropage_scpu01down
		#ply_a

		lda #%10111111
		bit Status
		bvs chrin_lastb
		bne nchrin_err_end
		#ldx_l_a buffer_input_exp
		gcc hard_chrin
;---------------------------------------
;BLK_RW_DIRTY, BLK_READ, BLK_RW
;
XIT_do_next	#xit_init

		#ldx_l_a buffer_currentp_exp
		lda buffer_info_exp,x
		pha
                jsr fileopprepare
                
		#ldy_l_a buffer_type_exp,x
                cpy #hi-lo
                bcs ill
                lda hi,y
                pha
                lda lo,y
                pha
                lda buffer_address_end_hi_exp,x
                cmp #1
                lda $a5                 ;buffer_address_hi_exp,x
                eor buffer_address_max_hi_exp,x
		rts

ill		brk

lo		.byte <(ill-1)
		.byte <(XIT_do_filenext-1)
		.byte <(XIT_do_dirnext-1)
		.byte <(XIT_do_direct-1)
                .byte <(XIT_do_pclink-1)
                .byte <(XIT_do_isofilenext-1)
                .byte <(XIT_do_ofsfilenext-1)
                .byte <(XIT_do_fatfilenext-1)

hi		.byte >(ill-1)
		.byte >(XIT_do_filenext-1)
		.byte >(XIT_do_dirnext-1)
		.byte >(XIT_do_direct-1)
                .byte >(XIT_do_pclink-1)
                .byte >(XIT_do_isofilenext-1)
                .byte >(XIT_do_ofsfilenext-1)
                .byte >(XIT_do_fatfilenext-1)
                
;---------------------------------------
XIT_do_filenext	.proc
		bcc dc
		bne XIT_do_pclink.eof	;eof mark?

		pla
		cmp #INFO_BLK_RW_DIRTY
		bne v1
		jsr XIT_flush_buf	;-=> error
		lda #INFO_BLK_RW
v1		pha

		jsr incfileposhi512

dc		jsr XIT_load_tree	;-=> error

		sec			;force clean
		jsr XIT_read_in_buffer	;-=> error
                
		jmp chrin_chrout_do_file_end
                .pend
;---------------------------------------
XIT_do_pclink	.proc
		bcc dc
eof		bne XIT_do_isofilenext.eof     	;eof mark?

dc              clc
		jsr pclink_chrinout
                pha
                jsr OP_init_buffer_pointers
                pla
                beq chrin_chrout_do_file_end.more
                tya
                bne *+4
                dec pom2
                dey
		lda pom2
                #jmps chrin_chrout_do_file_end.fer
		.pend
;---------------------------------------
XIT_do_isofilenext .proc
		bcc dc
eof		bne XIT_do_ofsfilenext.eof     	;eof mark?

		jsr incfileposhi2048

dc      	lda buffer_filepos2_exp,x
                sta next_sector+3
		lda buffer_filepos3_exp,x
                sta next_sector+2
                lda buffer_filepos4_exp,x

		ldy #3      
ry              lsr 
                ror next_sector+2
                ror next_sector+3
                dey
                bne ry
                sta next_sector+1
                sty next_sector

                txa
                asl
                asl
                tax
                ldy #3
ro              lda next_sector,y
                adc buffer_isofile_start_exp,x
                sta next_sector,y
                inx
                dey
                bpl ro

		jsr OP_buffer_start
                jsr XIT_read_a_sector

		#jmps chrin_chrout_do_file_end
		.pend
;---------------------------------------
XIT_do_ofsfilenext .proc
		bcc dc2
eof		bne XIT_do_fatfilenext.eof      	;eof mark?

dc2		
		lda #3
		sta buffer_address_low_exp,x
		lda buffer_start,x
                sta buffer_address_hi_exp,x
                jsr init_pom1_y

		lda #$3f
eos             #and_zp_y pom1
                ldx oldfsbyteorder,y
                sta next_sector,x
                lda #$ff
                iny
                cpy #4
                bcc eos

		lda pom2
                jsr XIT_read_a_sector

                jsr OP_buffer_start_pom1_y

                #lda_zp_y pom1
                asl
                asl
		lda buffer_address_max_hi_exp,x
                bcc ne
                iny
                #lda_zp_y pom1
                adc #3-1
		sta buffer_address_end_low_exp,x;Auto read first sector
                dey
                #lda0
                #adc_zp_y pom1
                and #$3f
                bne ne
                lda pom2
                ora #$80
                sta buffer_address_hi_exp,x
                #lda0
kf              adc pom2
ne		sta buffer_address_end_hi_exp,x

		#jmps chrin_chrout_do_file_end.be
		.pend
;---------------------------------------
XIT_do_fatfilenext .proc
		bcc dc
eof		bne read_file_eof      	;eof mark?

		jsr incfileposhi512
dc
		lda #>b_fat_current_exp
		jsr buffer_subpart_cp	;restore
                
                ldy #3
ir              #lda_zp_y current_path
                sta next_sector,y
                dey
                bpl ir
                ldy #4
                #lda_zp_y current_path
                jsr cluster2sector

		jsr OP_buffer_start
                jsr XIT_read_a_sector

		lda #>b_fat_current_exp
		jsr buffer_subpart_cp	;restore
                #ldx_l_a partition_currentp_exp;ok
                ldy #4
                #lda_zp_y current_path
                #adc1c
                #sta_zp_y current_path
                eor partition_fat_sectperclust_exp,x
                bne ok
                #sta_zp_y current_path

                ldy #3
ir2             #lda_zp_y current_path
                sta next_sector,y
                dey
                bpl ir2

                jsr XIT_nextcluster

		lda #>b_fat_current_exp
		jsr buffer_subpart_cp	;restore
                ldy #3
ir3             lda next_sector,y
                #sta_zp_y current_path
                dey
                bpl ir3

ok		#jmps chrin_chrout_do_file_end
		.pend
;---------------------------------------
; Common end of both chrin and chrout
;
read_file_eof  	.proc
		pla
		cmp #INFO_BLK_READ
		bne *+4
		lda #INFO_BLK_R_EOF
		cmp #INFO_BLK_RW
		bne *+4
		lda #INFO_BLK_RW_EOF
		cmp #INFO_BLK_RW_DIRTY
		bne *+4
		lda #INFO_BLK_RW_EOF_DIRTY
                pha
                lda #$40
		sta Status
		gne chrin_chrout_do_file_end.more
                .pend

chrin_chrout_do_file_end .proc
		jsr OP_init_buffer_pointers

		lda buffer_filesize1_exp,x
		sec
		sbc #1
		tay
		lda buffer_filesize2_exp,x
		sbc buffer_filepos2_exp,x
		sta pom2
		lda buffer_filesize3_exp,x
		sbc buffer_filepos3_exp,x
		bne more
		lda buffer_filesize4_exp,x
		sbc buffer_filepos4_exp,x
		bcc more
		bne more

		lda buffer_address_max_hi_exp,x
;		sec
		sbc buffer_start,x
		clc
		sbc pom2
		bcc more

		lda pom2
		ora buffer_start,x
		gcs fer

more		ldy #0
		lda buffer_address_max_hi_exp,x
fer		sta buffer_address_end_hi_exp,x
		tya
		sta buffer_address_end_low_exp,x
be
		pla			;be safe
		sta buffer_info_exp,x
		clc
		rts			;-=> ok
		.pend
;---------------------------------------
XIT_do_dirnext	.proc
		ldx #write_byte_r2end-write_byte_r2-1
htdo2		lda write_byte_r2,x
		sta start_zeropage,x
		dex
		bpl htdo2

		jsr OP_init_buffer_pointers

		lda buffer_start,x
		sta start_zeropage+6

		lda #>b_dir_misc_exp
		jsr buffer_subpart_cp

;                ldy #<b_dir_misc_exp
		ldx #direntry_backup_end_exp-direntry_backup_start_exp-1
nvw		#lda_zp_y current_path
		sta direntry_backup_start_exp,x
		iny
		dex
		bpl nvw
                
                #cnext direntry_next_sector_exp , next_sector

		lda direntry_first_exp
		bpl ndirlastbyte

		pla			;INFO_IGNORE
do_eof		lda #$40
		sta Status
		clc
		rts			;-=> last byte of file

ndirlastbyte	lda #>b_dir_searchstring_exp
		jsr buffer_subpart_cp	;restore searchstring

		lda buffer_start,x
                ora #>b_dir_limit_exp
		jsr XIT_dirlp2		;-=> error

		lda direntry_first_exp
		bpl hq
		lda start_zeropage+5
		bne *+4
		dec start_zeropage+6
		dec start_zeropage+5
hq
		#ldx_l_a buffer_currentp_exp
		lda start_zeropage+5
		sta buffer_address_end_low_exp,x
		lda start_zeropage+6
		sta buffer_address_end_hi_exp,x

		jsr next_sector_to_direntry_next_sector_exp

		lda #>b_dir_misc_exp
		jsr buffer_subpart_cp

;		ldy #<b_dir_misc_exp
		ldx #direntry_backup_end_exp-direntry_backup_start_exp-1
nvw2		lda direntry_backup_start_exp,x
		#sta_zp_y current_path
		iny
		dex
		bpl nvw2

		#ldx_l_a buffer_currentp_exp
		pla			;be safe
		sta buffer_info_exp,x
		clc			;-=> ok
		rts
                .pend
;---------------------------------------
XIT_do_direct	.proc
		jsr OP_init_buffer_pointers
		pla			;be safe
		sta buffer_info_exp,x
		jmp XIT_do_dirnext.do_eof
                .pend
;---------------------------------------


write_byte_r2	.proc
		.logical start_zeropage
		.if SUPERCPU=1
		txa			;1
		tyx			;2
		tay			;3
lp		pla			;4
		sta $1ff00,x		;8
		dex			;9
		bpl lp			;11
		tyx			;12
		rts			;13
		.else
		jmp lp			;3
lp		pla			;4
		sta $ff00,y		;7
		dey			;8
		bpl lp			;10
		rts			;11
		.fi
		.here
		.pend
write_byte_r2end

		.if (CHRIN_CHROUT_RAM=1) & (SUPERCPU=0)

chrin_data_sector_last__
		stx help_x_exp
		ldx buffer_input_exp
		lda newchrin_ram.tmpp+1
		sta $a4
		lda buffer_address_hi_exp,x
		jmp newchrin.chrin_data_sector_last

chrin_next_bank__
		stx help_x_exp
		ldx buffer_input_exp
		jmp newchrin.chrin_next_bank

chrin_not_data_sector__
		stx help_x_exp
		ldx buffer_input_exp
		jmp newchrin.chrin_not_data_sector

newchrin_ram_from
		.logical newchrin_ram_exp
newchrin_ram	.proc
		lda buffer_info_exp+STDIN;3
		bmi chrin_not_data_sector_;5

getin		asr #0			;7

		isb buffer_address_low_exp	;10
		eor #$ff		;12
		beq chrin_next_bank_	;14
		sta tmpp+1		;17
		lda buffer_address_hi_exp;20
		bmi chrin_data_sector_last_;22
		sta tmpp+2		;25

tmpp		lda $ffff		;28
		jmp exit2basic		;31

chrin_not_data_sector_
		jmp chrin_not_data_sector__;34
chrin_next_bank_
		jmp chrin_next_bank__	;37
chrin_data_sector_last_
		jmp chrin_data_sector_last__;40

newgetin_ram	.proc
		lda buffer_info_exp+STDIN;43
		bpl newchrin_ram.getin	;45
		jmp newgetin_back__	;48

newchrout_ram	.proc
		sta bce+1		;51
		lda buffer_info_exp+STDOUT;54
		asr #$01       		;56
		bcs chrout_not_data_sector_;58

		isb buffer_address_low_exp	;61
		eor #$ff		;63
		beq chrout_nxt_bank_	;65
		sta tmppp+1		;68
		lda buffer_address_hi_exp;71
		bmi chrout_data_sector_last_;73
		sta tmppp+2		;76

bce		lda #0			;78
tmppp		sta $ffff		;81
		jmp exit2basic		;84

chrout_not_data_sector_
		jmp chrout_not_data_sector__;87
chrout_nxt_bank_
		jmp chrout_nxt_bank__	;90
chrout_data_sector_last_
		jmp chrout_data_sector_last__;93
		.pend

		.here

newgetin_back__	stx help_x_exp
		ldx buffer_input_exp
		jmp newgetin.back

chrout_not_data_sector__
		stx help_x_exp
		ldx buffer_output_exp
		lda newchrout_ram.bce+1
		pha
		jmp newchrout.chrout_not_data_sector

chrout_nxt_bank__
		stx help_x_exp
		ldx buffer_output_exp
		lda newchrout_ram.bce+1
		pha
		lda #0
		jmp newchrout.chrout_nxt_bank

chrout_data_sector_last__
		stx help_x_exp
		ldx buffer_output_exp
		lda newchrout_ram.bce+1
		pha
		lda newchrout_ram.tmppp+1
		sta $a4
		lda buffer_address_hi_exp,x
		jmp newchrout.chrout_data_sector_last
		.fi
